@@ -1,9 +1,9 @@
 
 from .so4gp import DataGP
-from .so4gp import DataGP4clu
+from .so4gp import CluDataGP
 from .so4gp import GI
 from .so4gp import GP
-from .so4gp import GP4sw
+from .so4gp import ExtGP
 from .so4gp import NumericSS
 from .so4gp import TimeLag
 
@@ -20,7 +20,7 @@ from .so4gp import graank
 from .so4gp import psogps
 from .so4gp import rsgps
 
-from .so4gp import compare_gps
+from .so4gp import analyze_gps
 from .so4gp import get_num_cores
 from .so4gp import get_slurm_cores
 from .so4gp import write_file
