@@ -3,10 +3,10 @@
 @author: Dickson Owuor
 @credits: Thomas Runkler, Edmond Menya, and Anne Laurent
 @license: MIT
-@version: 0.2.5
+@version: 0.3.2
 @email: owuordickson@gmail.com
 @created: 21 July 2021
-@modified: 31 August 2022
+@modified: 07 October 2022
 
 SO4GP
 ------
@@ -35,6 +35,7 @@ SO4GP
 """
 
 import csv
+from collections import defaultdict
 from dateutil.parser import parse
 import time
 import gc
@@ -45,6 +46,8 @@ import multiprocessing as mp
 import os
 import pandas as pd
 import random
+import statistics
+from tabulate import tabulate
 from ypstruct import structure
 from sklearn.cluster import KMeans
 
@@ -127,6 +130,21 @@ class DataGP:
 
         no_bins: True if all none of the attr_cols yields a valid bitmap
 
+    The class provides the following function:
+        get_attr_cols: retrieves all the columns with data that is numeric and not date-tme
+
+        get_time_cols: retrieves the columns with date-time values
+
+        get_gi_bitmap: computes and returns the bitmap matrix corresponding to a GI
+
+        init_attributes: generates all the bitmap matrices of valid GIs
+
+        read (static): reads contents of a CSV file or data-frame
+
+        test_time (static): tests if a str represents a date-time variable
+
+        clean_data (static): cleans data (missing values, outliers) before extraction of GPs.
+
     """
 
     def __init__(self, data_source, min_sup=MIN_SUPPORT, eq=False):
@@ -211,6 +229,12 @@ class DataGP:
         return np.array(time_cols)
 
     def get_gi_bitmap(self, col):
+        """
+        Computes and returns the bitmap matrix corresponding to an attribute.
+
+        :param col: specific attribute (or column)
+        :return: numpy (bitmap)
+        """
         if col in self.time_cols:
             raise Exception("Error: " + str(self.titles[col][1].decode()) + " is a date/time column!")
         elif col >= self.col_count:
@@ -223,7 +247,7 @@ class DataGP:
                 temp_pos = np.where(col_data < col_data[:, np.newaxis], 1, 0)
             return temp_pos
 
-    def init_attributes(self, attr_data=None):
+    def init_bitmap(self, attr_data=None):
         """
         Generates bitmaps for columns with numeric objects. It only stores (attribute valid_bins) those bitmaps whose
         computed support values are greater or equal to the minimum support threshold value.
@@ -334,6 +358,12 @@ class DataGP:
 
     @staticmethod
     def test_time(date_str):
+        """
+        Tests if a str represents a date-time variable.
+
+        :param date_str: str value
+        :return: bool (True if it is date-time variable, False otherwise)
+        """
         # add all the possible formats
         try:
             if type(int(date_str)):
@@ -352,6 +382,11 @@ class DataGP:
 
     @staticmethod
     def clean_data(df):
+        """
+        Cleans a data-frame (i.e., missing values, outliers) before extraction of GPs
+        :param df: data-frame
+        :return: list (column titles), numpy (cleaned data)
+        """
         # 1. Remove objects with Null values
         df = df.dropna()
 
@@ -386,47 +421,202 @@ class DataGP:
         return titles, df.values
 
 
-class DataGP4clu(DataGP):
-    """Description of class DataGP4clu
-    A class for creating data-gp objects for the clustering approach. This class inherits the DataGP class which is used
-    to create data-gp objects. The classical data-gp object is meant to store all the parameters required by GP
-    algorithms to extract gradual patterns (GP). It takes a numeric file (in CSV format) as input and converts it into
-    an object whose attributes are used by algorithms to extract GPs.
+class DfsDataGP:
+
+    def __init__(self, file_path, min_sup=0):
+        self.thd_supp = min_sup
+        self.titles, self.data = DataGP.read(file_path)
+        self.row_count, self.col_count = self.data.shape
+        self.time_cols = self.get_time_cols()
+        self.attr_cols = self.get_attr_cols()
+        self.cost_matrix = np.ones((self.col_count, 3), dtype=int)
+        self.min_len = self.thd_supp * self.row_count
+        self.item_to_tids = defaultdict(set)
+        # self.encoded_data = np.array([])
+
+    def get_attr_cols(self):
+        """
+        Returns indices of all columns with non-datetime objects
+
+        :return: ndarray
+        """
+        all_cols = np.arange(self.col_count)
+        attr_cols = np.setdiff1d(all_cols, self.time_cols)
+        return attr_cols
+
+    def get_time_cols(self):
+        """
+        Tests each column's objects for date-time values. Returns indices of all columns with date-time objects
+
+        :return: ndarray
+        """
+        # Retrieve first column only
+        time_cols = list()
+        n = self.col_count
+        for i in range(n):  # check every column/attribute for time format
+            row_data = str(self.data[0][i])
+            try:
+                time_ok, t_stamp = DataGP.test_time(row_data)
+                if time_ok:
+                    time_cols.append(i)
+            except ValueError:
+                continue
+        return np.array(time_cols)
+
+    def encode_data(self):
+        attr_data = self.data.T
+        # self.attr_size = len(attr_data[self.attr_cols[0]])
+        size = self.row_count  # self.attr_size
+        n = len(self.attr_cols) + 2
+        encoded_data = list()
+        for i in range(size):
+            j = i + 1
+            if j >= size:
+                continue
+
+            temp_arr = np.empty([n, (size - j)], dtype=int)
+            temp_arr[0] = np.repeat(i, (size - j))
+            temp_arr[1] = np.arange(j, size)
+            k = 2
+            for col in self.attr_cols:
+                row_in = attr_data[col][i]
+                row_js = attr_data[col][(i+1):size]
+                v = col + 1
+                row = np.where(row_js > row_in, v, np.where(row_js < row_in, -v, 0))
+                temp_arr[k] = row
+                k += 1
+                pos_cost = np.count_nonzero(row == v)
+                neg_cost = np.count_nonzero(row == -v)
+                inv_cost = np.count_nonzero(row == 0)
+                self.cost_matrix[col][0] += (neg_cost + inv_cost)
+                self.cost_matrix[col][1] += (pos_cost + inv_cost)
+                self.cost_matrix[col][2] += (pos_cost + neg_cost)
+            temp_arr = temp_arr.T
+            encoded_data.extend(temp_arr)
+        self.data = None
+        gc.collect()
+        return np.array(encoded_data)
+
+    def remove_inv_attrs(self, encoded_data):
+        c_matrix = self.cost_matrix
+        # 1. remove invalid attributes
+        valid_a1 = list()
+        valid_a2 = [-2, -1]
+        for i in range(len(self.attr_cols)):
+            a = self.attr_cols[i]
+            valid = (c_matrix[a][0] < c_matrix[a][2]) or \
+                    (c_matrix[a][1] < c_matrix[a][2])
+            if valid:
+                valid_a1.append(i)
+                valid_a2.append(i)
+        self.attr_cols = self.attr_cols[valid_a1]
+        valid_a2 = np.array(valid_a2) + 2
+        encoded_data = encoded_data[:, valid_a2]
+        return encoded_data
+
+    def init_transaction_ids(self):
+
+        encoded_data = self.remove_inv_attrs(self.encode_data())
+
+        # 1. group similar items
+        for t in range(len(encoded_data)):
+            transaction = encoded_data[t][2:]
+            for item in transaction:
+                self.item_to_tids[item].add(tuple(encoded_data[t][:2]))
+
+        low_supp_items = [k for k, v in self.item_to_tids.items()
+                          if len(np.unique(np.array(list(v))[:, 0], axis=0))
+                          < self.min_len]
+        for item in low_supp_items:
+            del self.item_to_tids[item]
+        gc.collect()
+
+
+class CluDataGP(DataGP):
+    """Description of class CluDataGP (Clustering DataGP)
+
+    CluDataGP stands for Clustering DataGP. It is a class that inherits the DataGP class in order to create data-gp
+    objects for the clustering approach. This class inherits the DataGP class which is used to create data-gp objects.
+    The classical data-gp object is meant to store all the parameters required by GP algorithms to extract gradual
+    patterns (GP). It takes a numeric file (in CSV format) as input and converts it into an object whose attributes are
+    used by algorithms to extract GPs.
+
+    class DataGP provides the following attributes:
+        thd_supp: minimum support threshold
+
+        equal: eq value
+
+        titles: column names of data source
+
+        data: all the objects organized into their respective column
+
+        row_count: number of objects
+
+        col_count: number of all columns
+
+        time_cols: column indices of the columns with data-time objects
+
+        attr_cols: column indices of the columns with numeric values
+
+        valid_bins: valid bitmaps (in the form of ndarray) of all gradual items corresponding to the attr_cols, a bitmap
+        is valid if its computed support is equal or greater than the minimum support threshold
+
+        no_bins: True if all none of the attr_cols yields a valid bitmap
 
     This class adds the parameters required for clustering gradual items to the data-gp object. The class provides the
     following additional attributes:
+        e_prob: erasure probability (a value between 0 - 1)
 
-    e_prob: erasure probability (a value between 0 - 1)
+        mat_iter: maximum iteration value for score vector estimation
 
-    mat_iter: maximum iteration value for score vector estimation
+    CluDataGP adds the following funtions:
+        construct_matrices: generates the net-win matrix
+
+        infer_gps: infers GPs from clusters of Gradual Items
+
+        estimate_score_vector: estimates the score vector based on the cumulative wins
+
+        estimate_support:  estimates the frequency support of a GP based on its score vector
 
     """
 
-    def __init__(self, *args, e_prob=ERASURE_PROBABILITY, max_iter=SCORE_VECTOR_ITERATIONS):
-        """Description of class DataGP4clu
+    def __init__(self, *args, e_prob=ERASURE_PROBABILITY, max_iter=SCORE_VECTOR_ITERATIONS, all=False):
+        """Description of class CluDataGP (Clustering DataGP)
+
         A class for creating data-gp objects for the clustering approach. This class inherits the DataGP class which is
         used to create data-gp objects. This class adds the parameters required for clustering gradual items to the
-        data-gp object. The class provides the following additional attributes:
+        data-gp object.
 
-        e_prob: erasure probability (a value between 0 - 1). Erasure probability determines the proportion of ij columns
-        to be used by the algorithm, the rest are ignored.
+        The class provides the following additional attributes:
+            e_prob: erasure probability (a value between 0 - 1). Erasure probability determines the proportion of ij columns
+            to be used by the algorithm, the rest are ignored.
 
-        mat_iter: maximum iteration value for score vector estimation
+            mat_iter: maximum iteration value for score vector estimation
 
         :param args: [required] data-source, [optional] minimum-support
         :param e_prob: [optional] erasure probability, the default is 0.5
         :param max_iter: [optional] maximum iteration for score vector estimation, the default is 10
         """
-        super(DataGP4clu, self).__init__(*args)
+        super(CluDataGP, self).__init__(*args)
         self.erasure_probability = e_prob
         """:type erasure_probability: float"""
         self.max_iteration = max_iter
         """:type max_iteration: int"""
-        self.gradual_items, self.cum_wins, self.net_win_mat, self.ij = self.construct_matrices(e_prob)
-        """:type gradual_items: ndarray"""
-        """:type cum_wins: ndarray"""
-        """:type net_win_mat: ndarray"""
-        """:type ij: ndarray"""
+        if not all:
+            self.gradual_items, self.cum_wins, self.net_win_mat, self.ij = self.construct_matrices(e_prob)
+            """:type gradual_items: ndarray"""
+            """:type cum_wins: ndarray"""
+            """:type net_win_mat: ndarray"""
+            """:type ij: ndarray"""
+            self.win_mat = np.array([])
+            """:type win_mat: ndarray"""
+        else:
+            self.gradual_items, self.win_mat, self.cum_wins, self.net_win_mat, self.ij = self.construct_all_matrices()
+            """:type gradual_items: ndarray"""
+            """:type win_mat: ndarray"""
+            """:type cum_wins: ndarray"""
+            """:type net_win_mat: ndarray"""
+            """:type ij: ndarray"""
 
     def construct_matrices(self, e):
         """
@@ -439,13 +629,20 @@ class DataGP4clu(DataGP):
         n = self.row_count
         prob = 1 - e  # Sample probability
 
-        # 1a. Generate random pairs using erasure-probability
-        total_pair_count = int(n * (n - 1) * 0.5)
-        rand_1d = np.random.choice(n, int(prob * total_pair_count) * 2, replace=True)
-        pair_ij = np.reshape(rand_1d, (-1, 2))
+        if prob == 1:
+            # 1a. Generate all possible pairs
+            pair_ij = np.array(np.meshgrid(np.arange(n), np.arange(n))).T.reshape(-1, 2)
 
-        # 1b. Remove duplicates
-        pair_ij = pair_ij[np.argwhere(pair_ij[:, 0] != pair_ij[:, 1])[:, 0]]
+            # 1b. Remove duplicates or reversed pairs
+            pair_ij = pair_ij[np.argwhere(pair_ij[:, 0] < pair_ij[:, 1])[:, 0]]
+        else:
+            # 1a. Generate random pairs using erasure-probability
+            total_pair_count = int(n * (n - 1) * 0.5)
+            rand_1d = np.random.choice(n, int(prob * total_pair_count) * 2, replace=True)
+            pair_ij = np.reshape(rand_1d, (-1, 2))
+
+            # 1b. Remove duplicates
+            pair_ij = pair_ij[np.argwhere(pair_ij[:, 0] != pair_ij[:, 1])[:, 0]]
 
         # 2. Variable declarations
         attr_data = self.data.T  # Feature data objects
@@ -453,6 +650,64 @@ class DataGP4clu(DataGP):
         s_mat = []  # S-Matrix (made up of S-Vectors)
         cum_wins = []  # Cumulative wins
 
+        # 3. Construct S matrix from data set
+        for col in np.nditer(self.attr_cols):
+            # Feature data objects
+            col_data = np.array(attr_data[col], dtype=np.float)  # Feature data objects
+
+            # Cumulative Wins: for estimation of score-vector
+            temp_cum_wins = np.where(col_data[pair_ij[:, 0]] < col_data[pair_ij[:, 1]], 1,
+                                     np.where(col_data[pair_ij[:, 0]] > col_data[pair_ij[:, 1]], -1, 0))
+            # print(col)
+            # print(temp_cum_wins)
+
+            # S-vector
+            s_vec = np.zeros((n,), dtype=np.int32)
+            for w in [1, -1]:
+                positions = np.flatnonzero(temp_cum_wins == w)
+                i, counts_i = np.unique(pair_ij[positions, 0], return_counts=True)
+                j, counts_j = np.unique(pair_ij[positions, 1], return_counts=True)
+                s_vec[i] += w * counts_i  # i wins/loses (1/-1)
+                s_vec[j] += -w * counts_j  # j loses/wins (1/-1)
+            # print(s_vec)
+            # print("\n")
+            # Normalize S-vector
+            if np.count_nonzero(s_vec) > 0:
+                s_vec[s_vec > 0] = 1  # Normalize net wins
+                s_vec[s_vec < 0] = -1  # Normalize net loses
+
+                lst_gis.append(GI(col, '+'))
+                cum_wins.append(temp_cum_wins)
+                s_mat.append(s_vec)
+
+                lst_gis.append(GI(col, '-'))
+                cum_wins.append(-temp_cum_wins)
+                s_mat.append(-s_vec)
+
+        return np.array(lst_gis), np.array(cum_wins), np.array(s_mat), pair_ij
+
+    def construct_all_matrices(self):
+        """
+        Generates all the gradual items and, constructs: (1) win matrix (2) net-win matrix, (3) cumulative wins, (4) pairwise objects.
+
+        :return: list of gradual items, win matrix, net-win matrix, cumulative wins matrix, selected pairwise (ij) objects
+        """
+
+        n = self.row_count
+
+        # 1a. Generate all possible pairs
+        pair_ij = np.array(np.meshgrid(np.arange(n), np.arange(n))).T.reshape(-1, 2)
+
+        # 1b. Remove duplicates or reversed pairs
+        pair_ij = pair_ij[np.argwhere(pair_ij[:, 0] < pair_ij[:, 1])[:, 0]]
+
+        # 2. Variable declarations
+        attr_data = self.data.T  # Feature data objects
+        lst_gis = []  # List of GIs
+        s_mat = []  # S-Matrix (made up of S-Vectors)
+        w_mat = []  # win matrix
+        cum_wins = []  # Cumulative wins
+
         # 3. Construct S matrix from data set
         for col in np.nditer(self.attr_cols):
             # Feature data objects
@@ -473,6 +728,8 @@ class DataGP4clu(DataGP):
 
             # Normalize S-vector
             if np.count_nonzero(s_vec) > 0:
+                w_mat.append(np.copy(s_vec))
+
                 s_vec[s_vec > 0] = 1  # Normalize net wins
                 s_vec[s_vec < 0] = -1  # Normalize net loses
 
@@ -484,11 +741,11 @@ class DataGP4clu(DataGP):
                 cum_wins.append(-temp_cum_wins)
                 s_mat.append(-s_vec)
 
-        return np.array(lst_gis), np.array(cum_wins), np.array(s_mat), pair_ij
+        return np.array(lst_gis), np.array(w_mat), np.array(cum_wins), np.array(s_mat), pair_ij
 
     def infer_gps(self, clusters):
         """
-        A function that infers GPs from a clusters of gradual items.
+        A function that infers GPs from clusters of gradual items.
 
         :param clusters: [required] groups of gradual items clustered through K-MEANS algorithm
         :return: list of (str) patterns, list of GP objects
@@ -518,7 +775,7 @@ class DataGP4clu(DataGP):
 
                 # 4. Infer GPs from the clusters
                 if est_sup >= self.thd_supp:
-                    gp = GP4sw()
+                    gp = ExtGP()
                     for gi in cluster_gis:
                         gp.add_gradual_item(gi)
                     gp.set_support(est_sup)
@@ -588,6 +845,10 @@ class DataGP4clu(DataGP):
 # -------- OTHER METHODS -----------
 
 def get_num_cores():
+    """
+    Finds the count of CPU cores in a computer or a SLURM super-computer.
+    :return: number of cpu cores (int)
+    """
     num_cores = get_slurm_cores()
     if not num_cores:
         num_cores = mp.cpu_count()
@@ -595,6 +856,10 @@ def get_num_cores():
 
 
 def get_slurm_cores():
+    """
+    Test computer to see if it is a SLURM environment, then gets number of CPU cores.
+    :return: count of CPUs (int) or False
+    """
     try:
         cores = int(os.environ['SLURM_JOB_CPUS_PER_NODE'])
         return cores
@@ -616,6 +881,13 @@ def get_slurm_cores():
 
 
 def write_file(data, path, wr=True):
+    """
+    Writes data into a file
+    :param data: information to be written
+    :param path: name of file and storage path
+    :param wr: writes data into file if True
+    :return:
+    """
     if wr:
         with open(path, 'w') as f:
             f.write(data)
@@ -624,22 +896,45 @@ def write_file(data, path, wr=True):
         pass
 
 
-def compare_gps(file, min_sup,  est_gps):
-    d_set = DataGP(file, min_sup)
-    d_set.init_attributes()
-    wr_line = "\n\nComparison : Estimated Support, True Support" + '\n'
+def analyze_gps(file, min_sup, est_gps, approach='bfs'):
+    """
+    For each estimated GP, computes its true support using GRAANK approach and returns the statistics (% error, and standard deviation)
+    :param file: data set file
+    :param min_sup: minimum support (set by user)
+    :param est_gps: estimated GPs
+    :param approach: 'bfs' (default) or 'dfs'
+    :return: tabulated results
+    """
+    if approach == 'dfs':
+        d_set = DfsDataGP(file, min_sup)
+        d_set.init_transaction_ids()
+    else:
+        d_set = DataGP(file, min_sup)
+        d_set.init_bitmap()
+    headers = ["Gradual Pattern", "Estimated Support", "True Support", "Percentage Error", "Standard Deviation"]
+    data = []
     for est_gp in est_gps:
         est_sup = est_gp.support
         est_gp.set_support(0)
-        true_gp = est_gp.validate(d_set)
-        true_sup = true_gp.support
-        if true_sup == 0:
-            true_sup = -1
-        if len(true_gp.gradual_items) == len(est_gp.gradual_items):
-            wr_line += (str(est_gp.to_string()) + ' : ' + str(round(est_sup, 3)) + ', ' + str(round(true_sup, 3)) + '\n')
+        if approach == 'dfs':
+            true_gp = est_gp.validate_tree(d_set)
         else:
-            wr_line += (str(est_gp.to_string()) + ' : ' + str(round(est_sup, 3)) + ', -1\n')
-    return wr_line
+            true_gp = est_gp.validate_graank(d_set)
+        true_sup = true_gp.support
+
+        if true_sup == 0:
+            percentage_error = np.inf
+            st_dev = np.inf
+        else:
+            percentage_error = ((est_sup - true_sup) / true_sup) * 100
+            st_dev = statistics.stdev([est_sup, true_sup])
+
+        if len(true_gp.gradual_items) == len(est_gp.gradual_items):
+            data.append([est_gp.to_string(), round(est_sup, 3), round(true_sup, 3), str(round(percentage_error, 3))+'%',
+                         round(st_dev, 3)])
+        else:
+            data.append([est_gp.to_string(), round(est_sup, 3), -1, np.inf, np.inf])
+    return tabulate(data, headers=headers)
 
 
 # -------- GRADUAL PATTERNS -------------
@@ -704,7 +999,7 @@ class GI:
     def inv(self):
         """
         Inverts a GI to the opposite variation (i.e., from - to +; or, from + to -)
-        :return: inverted GI
+        :return: inverted GI (ndarray)
         """
         if self.symbol == '+':
             # temp = tuple([self.attribute_col, '-'])
@@ -716,6 +1011,18 @@ class GI:
             temp = np.array((self.attribute_col, 'x'), dtype='i, S1')
         return temp
 
+    def inv_gi(self):
+        """
+        Inverts a GI to the opposite variation (i.e., from - to +; or, from + to -)
+        :return: inverted GI object
+        """
+        if self.symbol == '+':
+            sym = '-'
+        else:
+            sym = '+'
+        new_gi = GI(self.attribute_col, sym)
+        return new_gi
+
     def as_integer(self):
         """
         Converts variation symbol into an integer (i.e., + to 1; and - to -1)
@@ -785,6 +1092,7 @@ class GP:
 
      The class has the following attributes:
         gradual_items: list if GIs
+
         support: computed support value as a float
 
     """
@@ -799,6 +1107,7 @@ class GP:
 
              The class has the following attributes:
                 gradual_items: list if GIs
+
                 support: computed support value as a float
 
             """
@@ -825,6 +1134,19 @@ class GP:
         else:
             pass
 
+    def add_items_from_list(self, lst_items):
+        """
+        Adds gradual items from a list of str or a list of sets.
+
+        :param lst_items: str or set
+        :return: none
+        """
+        for str_gi in lst_items:
+            if type(str_gi[1]) is str:
+                self.add_gradual_item(GI(int(str_gi[0]), str_gi[1]))
+            elif type(str_gi[1]) is bytes:
+                self.add_gradual_item(GI(int(str_gi[0]), str(str_gi[1].decode())))
+
     def get_pattern(self):
         """
         Returns the gradual pattern (GP) as a list
@@ -968,26 +1290,43 @@ class GP:
         return [pattern, self.support]
 
 
-class GP4sw(GP):
-    """Description of class GP4sw (Gradual Pattern for swarm)
+class ExtGP(GP):
+    """Description of class ExtGP (Extended Gradual Pattern)
 
-    A class that inherits class GP which is used to create GP objects. a GP object is a set of gradual items and its
-    quality is measured by its computed support value. For example given a data set with 3 columns (age, salary, cars)
-    and 10 objects. A GP may take the form: {age+, salary-} with a support of 0.8. This implies that 8 out of 10 objects
-    have the values of column age 'increasing' and column 'salary' decreasing.
+    A class that inherits class GP which is used to create more capable GP objects. a GP object is a set of gradual
+    items and its quality is measured by its computed support value. For example given a data set with 3 columns
+    (age, salary, cars) and 10 objects. A GP may take the form: {age+, salary-} with a support of 0.8. This implies that
+    8 out of 10 objects have the values of column age 'increasing' and column 'salary' decreasing.
 
     The class GP has the following attributes:
         gradual_items: list if GIs
+
         support: computed support value as a float
 
-    The class GP4sw adds the following functions:
+    The class ExtGP adds the following functions:
         validate: used to validate GPs
+
         check_am: used to verify if a GP obeys anti-monotonicity
+
         is_duplicate: checks a GP is already extracted
 
     """
 
-    def validate(self, d_set):
+    def __init__(self):
+        """Description of class ExtGP (Extended Gradual Pattern)
+
+        A class that inherits class GP which is used to create more powerful GP objects that can be used in mining
+        approaches that implement swarm optimization techniques or cluster analysis or classification algorithms.
+
+        It adds the following attribute:
+            freq_count: frequency count of a particular GP object.
+
+        """
+        super(ExtGP, self).__init__()
+        self.freq_count = 0
+        """:type freq_count: int"""
+
+    def validate_graank(self, d_set):
         """
         Validates a candidate gradual pattern (GP) based on support computation. A GP is invalid if its support value is
         less than the minimum support threshold set by the user.
@@ -998,8 +1337,8 @@ class GP4sw(GP):
         # pattern = [('2', '+'), ('4', '+')]
         min_supp = d_set.thd_supp
         n = d_set.attr_size
-        gen_pattern = GP4sw()
-        """type gen_pattern: GP"""
+        gen_pattern = ExtGP()
+        """type gen_pattern: ExtGP"""
         bin_arr = np.array([])
 
         for gi in self.gradual_items:
@@ -1023,6 +1362,57 @@ class GP4sw(GP):
         else:
             return gen_pattern
 
+    def validate_tree(self, d_set):
+        min_supp = d_set.thd_supp
+        n = d_set.row_count
+        gen_pattern = ExtGP()
+        """type gen_pattern: ExtGP"""
+        temp_tids = None
+        for gi in self.gradual_items:
+            gi_int = gi.as_integer()
+            node = int(gi_int[0] + 1) * gi_int[1]
+            gi_int = (gi.inv_gi()).as_integer()
+            node_inv = int(gi_int[0] + 1) * gi_int[1]
+            for k, v in d_set.item_to_tids.items():
+                if node == k:
+                    if temp_tids is None:
+                        temp_tids = v
+                        gen_pattern.add_gradual_item(gi)
+                    else:
+                        temp = temp_tids.copy()
+                        temp = temp.intersection(v)
+                        if len(temp) > 1:
+                            x = np.unique(np.array(list(temp))[:, 0], axis=0)
+                            supp = len(x) / n
+                        else:
+                            supp = len(temp) / n
+
+                        if supp >= min_supp:
+                            temp_tids = temp.copy()
+                            gen_pattern.add_gradual_item(gi)
+                            gen_pattern.set_support(supp)
+                elif node_inv == k:
+                    if temp_tids is None:
+                        temp_tids = v
+                        gen_pattern.add_gradual_item(gi)
+                    else:
+                        temp = temp_tids.copy()
+                        temp = temp.intersection(v)
+                        if len(temp) > 1:
+                            x = np.unique(np.array(list(temp))[:, 0], axis=0)
+                            supp = len(x) / n
+                        else:
+                            supp = len(temp) / n
+
+                        if supp >= min_supp:
+                            temp_tids = temp.copy()
+                            gen_pattern.add_gradual_item(gi)
+                            gen_pattern.set_support(supp)
+        if len(gen_pattern.gradual_items) <= 1:
+            return self
+        else:
+            return gen_pattern
+
     def check_am(self, gp_list, subset=True):
         """
         Anti-monotonicity check. Checks if a GP is a subset or superset of an already existing GP
@@ -1149,8 +1539,8 @@ class NumericSS:
         :return: GP that is decoded from the position value
         """
 
-        temp_gp = GP4sw()
-        ":type temp_gp: GP4sw"
+        temp_gp = ExtGP()
+        ":type temp_gp: ExtGP"
         if position is None:
             return temp_gp
 
@@ -1299,7 +1689,7 @@ def graank(f_path=None, min_sup=MIN_SUPPORT, eq=False, return_gps=False):
 
     d_set = DataGP(f_path, min_sup, eq)
     """:type d_set: DataGP"""
-    d_set.init_attributes()
+    d_set.init_bitmap()
 
     patterns = []
     """:type patterns: GP list"""
@@ -1409,7 +1799,7 @@ def acogps(f_path, min_supp=MIN_SUPPORT, evaporation_factor=EVAPORATION_FACTOR,
     # 0. Initialize and prepare data set
     d_set = DataGP(f_path, min_supp)
     """:type d_set: DataGP"""
-    d_set.init_attributes()
+    d_set.init_bitmap()
     # attr_index = d_set.attr_cols
     # e_factor = evaporation_factor
     d, attr_keys = gen_d(d_set.valid_bins)  # distance matrix (d) & attributes corresponding to d
@@ -1447,8 +1837,8 @@ def acogps(f_path, min_supp=MIN_SUPPORT, evaporation_factor=EVAPORATION_FACTOR,
                 is_sub = rand_gp.check_am(winner_gps, subset=True)
                 if is_super or is_sub:
                     continue
-                gen_gp = rand_gp.validate(d_set)
-                """:type gen_gp: GP4sw"""
+                gen_gp = rand_gp.validate_graank(d_set)
+                """:type gen_gp: ExtGP"""
                 is_present = gen_gp.is_duplicate(winner_gps, loser_gps)
                 is_sub = gen_gp.check_am(winner_gps, subset=True)
                 if is_present or is_sub:
@@ -1484,8 +1874,8 @@ def acogps(f_path, min_supp=MIN_SUPPORT, evaporation_factor=EVAPORATION_FACTOR,
 
 def gen_aco_candidates(attr_keys, d, p_matrix, e_factor):
     v_matrix = d
-    pattern = GP4sw()
-    ":type pattern: GP4sw"
+    pattern = ExtGP()
+    ":type pattern: ExtGP"
 
     # 1. Generate gradual items with the highest pheromone and visibility
     m = p_matrix.shape[0]
@@ -1581,7 +1971,7 @@ def gagps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, n_pop=N_
 
     # Prepare data set
     d_set = DataGP(data_src, min_supp)
-    d_set.init_attributes()
+    d_set.init_bitmap()
     attr_keys = [GI(x[0], x[1].decode()).as_string() for x in d_set.valid_bins[:, 0]]
 
     if d_set.no_bins:
@@ -1699,8 +2089,8 @@ def gagps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, n_pop=N_
         pop = sorted(pop, key=lambda x: x.cost)
         pop = pop[0:n_pop]
 
-        best_gp = NumericSS.decode_gp(attr_keys, best_sol.position).validate(d_set)
-        """:type best_gp: GP4sw"""
+        best_gp = NumericSS.decode_gp(attr_keys, best_sol.position).validate_graank(d_set)
+        """:type best_gp: ExtGP"""
         is_present = best_gp.is_duplicate(best_patterns)
         is_sub = best_gp.check_am(best_patterns, subset=True)
         if is_present or is_sub:
@@ -1827,7 +2217,7 @@ def psogps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, n_parti
     """
     # Prepare data set
     d_set = DataGP(data_src, min_supp)
-    d_set.init_attributes()
+    d_set.init_bitmap()
     # self.target = 1
     # self.target_error = 1e-6
     attr_keys = [GI(x[0], x[1].decode()).as_string() for x in d_set.valid_bins[:, 0]]
@@ -1902,8 +2292,8 @@ def psogps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, n_parti
                            (coef_g * random.random()) * (gbest_particle.position - particle_pop[i].position)
             particle_pop[i].position = particle_pop[i].position + new_velocity
 
-        best_gp = NumericSS.decode_gp(attr_keys, best_particle.position).validate(d_set)
-        """:type best_gp: GP4sw"""
+        best_gp = NumericSS.decode_gp(attr_keys, best_particle.position).validate_graank(d_set)
+        """:type best_gp: ExtGP"""
         is_present = best_gp.is_duplicate(best_patterns)
         is_sub = best_gp.check_am(best_patterns, subset=True)
         if is_present or is_sub:
@@ -1978,7 +2368,7 @@ def hcgps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, step_siz
     """
     # Prepare data set
     d_set = DataGP(data_src, min_supp)
-    d_set.init_attributes()
+    d_set.init_bitmap()
     attr_keys = [GI(x[0], x[1].decode()).as_string() for x in d_set.valid_bins[:, 0]]
 
     if d_set.no_bins:
@@ -2031,8 +2421,8 @@ def hcgps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, step_siz
         eval_count += 1
         str_eval += "{}: {} \n".format(eval_count, best_sol.cost)
 
-        best_gp = NumericSS.decode_gp(attr_keys, best_sol.position).validate(d_set)
-        """:type best_gp: GP4sw"""
+        best_gp = NumericSS.decode_gp(attr_keys, best_sol.position).validate_graank(d_set)
+        """:type best_gp: ExtGP"""
         is_present = best_gp.is_duplicate(best_patterns)
         is_sub = best_gp.check_am(best_patterns, subset=True)
         if is_present or is_sub:
@@ -2104,7 +2494,7 @@ def rsgps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, return_g
     """
     # Prepare data set
     d_set = DataGP(data_src, min_supp)
-    d_set.init_attributes()
+    d_set.init_bitmap()
     attr_keys = [GI(x[0], x[1].decode()).as_string() for x in d_set.valid_bins[:, 0]]
 
     if d_set.no_bins:
@@ -2151,8 +2541,8 @@ def rsgps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, return_g
         eval_count += 1
         str_eval += "{}: {} \n".format(eval_count, best_sol.cost)
 
-        best_gp = NumericSS.decode_gp(attr_keys, best_sol.position).validate(d_set)
-        """:type best_gp: GP4sw"""
+        best_gp = NumericSS.decode_gp(attr_keys, best_sol.position).validate_graank(d_set)
+        """:type best_gp: ExtGP"""
         is_present = best_gp.is_duplicate(best_patterns)
         is_sub = best_gp.check_am(best_patterns, subset=True)
         if is_present or is_sub:
@@ -2197,7 +2587,7 @@ def rsgps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, return_g
 @created: 01 March 2022
 @modified: 25 April 2022
 CHANGES:
-1. Uses class DataGP4clu
+1. Uses class CluDataGP
 """
 
 
@@ -2232,8 +2622,8 @@ def clugps(data_src, min_supp=MIN_SUPPORT, e_probability=ERASURE_PROBABILITY, sv
     """
 
     # 1. Create a DataGP object
-    d_gp = DataGP4clu(data_src, min_supp, e_prob=e_probability, max_iter=sv_max_iter)
-    """:type d_gp: DataGP4clu"""
+    d_gp = CluDataGP(data_src, min_supp, e_prob=e_probability, max_iter=sv_max_iter)
+    """:type d_gp: CluDataGP"""
 
     # 2. Generate net-win matrices
     s_matrix = d_gp.net_win_mat  # Net-win matrix (S)
